==> syd-3.21.0.tar.gz: syd-v3.21.0/src/syd-elf.rs <==
//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-elf.rs: syd's secure ldd(1) alternative
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    io::{Read, Seek, SeekFrom},
    path::{Path, PathBuf},
    process::ExitCode,
};

use anyhow::{bail, Context, Result};
use getargs::{Opt, Options};
use goblin::elf::{
    header::{ET_CORE, ET_DYN, ET_EXEC, ET_NONE, ET_REL},
    Elf,
};
use nix::unistd::Pid;
use regex::Regex;

fn main() -> Result<ExitCode> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = std::env::args().skip(1).collect::<Vec<_>>();

    // If no arguments are provided, display help
    if args.is_empty() {
        help();
        return Ok(ExitCode::SUCCESS);
    }

    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut root = None;
    let mut addr = None;
    let mut name = None;
    let mut list = false;
    let mut etyp = false;
    let mut ever = false;
    let mut build_id = false;
    let mut is_32bit = false;
    let mut is_64bit = false;
    let mut is_dynamic = false;
    let mut is_static = false;

    // SAFETY: We panic on parse errors.
    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('3') => {
                is_32bit = true;
            }
            Opt::Short('6') => {
                is_64bit = true;
            }
            Opt::Short('d') => {
                is_dynamic = true;
            }
            Opt::Short('s') => {
                is_static = true;
            }
            Opt::Short('b') => {
                build_id = true;
            }
            Opt::Short('l') => {
                list = true;
            }
            Opt::Short('t') => {
                etyp = true;
            }
            Opt::Short('v') => {
                ever = true;
            }
            Opt::Short('r') => {
                addr = match opts.value() {
                    Ok(value) => {
                        match u64::from_str_radix(
                            value.to_ascii_lowercase().trim_start_matches("0x"),
                            16,
                        ) {
                            Ok(value) => Some(value),
                            Err(_) => bail!("Invalid address specified for -r!"),
                        }
                    }
                    Err(_) => bail!("-r requires an argument!"),
                }
            }
            Opt::Short('n') => {
                name = match opts.value() {
                    Ok(value) => match Regex::new(value) {
                        Ok(value) => Some(value),
                        Err(_) => bail!("Invalid regular expression specified for -n!"),
                    },
                    Err(_) => bail!("-n requires an argument!"),
                }
            }
            Opt::Short('R') => {
                root = match opts.value() {
                    Ok(value) => Some(PathBuf::from(value)),
                    Err(_) => bail!("-R requires an argument!"),
                }
            }
            _ => bail!("Unknown option: {opt:?}!"),
        }
    }

    let flags = [
        build_id,
        is_32bit,
        is_64bit,
        is_dynamic,
        is_static,
        list,
        addr.is_some(),
        name.is_some(),
        etyp,
        ever,
    ];
    let parse = match flags.iter().filter(|&&flag| flag).count() {
        0 => false,
        1 => true,
        _ => bail!("At most one of -3, -6, -b, -d, -l, -n, -r, -s, t, and -v must be given!"),
    };

    let path = match opts.positionals().next() {
        None => bail!("Expected an ELF file as the first argument!"),
        Some(value) => Path::new(value)
            .canonicalize()
            .context("Failed to canonicalize argument!")?,
    };

    let pid = Pid::this();
    let (mut exe, size) =
        syd::fs::safe_open_file(&path).context("Failed to open file for reading!")?;
    let prog =
        syd::fs::binfmt_script(&mut exe, pid, true).context("Failed to determine interpreter")?;

    if parse {
        let bytes = if let Some(prog) = prog {
            if etyp {
                println!("script");
                return Ok(ExitCode::SUCCESS);
            }
            drop(exe);
            std::fs::read(prog).context("Failed to read ELF file")?
        } else {
            exe.seek(SeekFrom::Start(0))
                .context("Failed to seek ELF file")?;

            let mut bytes = Vec::new();
            exe.read_to_end(&mut bytes)
                .context("Failed to read ELF file")?;
            drop(exe);
            bytes
        };
        let elf = Elf::parse(&bytes).context("Failed to parse ELF file")?;

        if is_32bit {
            return Ok(if !elf.is_64 {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if is_64bit {
            return Ok(if elf.is_64 {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if is_dynamic {
            return Ok(if elf.dynamic.is_some() {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if is_static {
            return Ok(if elf.dynamic.is_none() {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if build_id {
            // Use `iter_note_headers` to iterate over PT_NOTE program headers.
            if let Some(notes) = elf.iter_note_headers(&bytes) {
                for note in notes.flatten() {
                    if note.name == "GNU" && note.n_type == 3 {
                        println!("{}", const_hex::encode(note.desc));
                        return Ok(ExitCode::SUCCESS);
                    }
                }
            }
            return Ok(ExitCode::FAILURE);
        } else if list {
            for sym in &elf.syms {
                if let Some(name) = elf.strtab.get_at(sym.st_name) {
                    println!("{:#x}:{name}", sym.st_value);
                }
            }
            return Ok(ExitCode::SUCCESS);
        } else if etyp {
            let name = match elf.header.e_type {
                ET_NONE => "none".to_string(),
                ET_REL => "rel".to_string(),
                ET_EXEC => "exe".to_string(),
                ET_DYN => "obj".to_string(),
                ET_CORE => "core".to_string(),
                _ => format!("{:#x}", elf.header.e_type),
            };
            println!("{name}");
            return Ok(ExitCode::SUCCESS);
        } else if ever {
            println!("{}", elf.header.e_version);
            return Ok(ExitCode::SUCCESS);
        } else if let Some(addr) = addr {
            let name = elf
                .syms
                .iter()
                .find(|sym| sym.st_value == addr)
                .and_then(|sym| elf.strtab.get_at(sym.st_name));
            return Ok(if let Some(name) = name {
                println!("{addr:#x}:{name}");
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if let Some(regex) = name {
            let mut is_match = false;
            for sym in &elf.syms {
                if let Some(name) = elf.strtab.get_at(sym.st_name) {
                    if regex.is_match(name) {
                        is_match = true;
                        println!("{:#x}:{name}", sym.st_value);
                    }
                }
            }
            return Ok(if is_match {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else {
            unreachable!();
        }
    }

    let deps = if let Some(ref p) = prog {
        drop(exe); // no longer necessary.

        let mut deps = vec![p.to_path_buf(), path];
        let (mut exe, size) = syd::fs::safe_open_file(p).context("Failed to open interpreter")?;
        syd::fs::ldd(
            &mut exe, size, &mut deps, root, pid, false, false, false, true,
        )
        .context("Failed to analyze interpreter")?;
        drop(exe);

        deps
    } else {
        exe.seek(SeekFrom::Start(0))
            .context("Failed to seek ELF file")?;

        let mut deps = vec![path];
        syd::fs::ldd(
            &mut exe, size, &mut deps, root, pid, false, false, false, true,
        )
        .context("Failed to analyze ELF file")?;
        drop(exe);

        deps
    };

    for path in deps {
        println!("{}", path.display());
    }

    Ok(ExitCode::SUCCESS)
}

fn help() {
    println!("Usage: syd-elf [-36bdhlrstv] [-R root] binary|script");
    println!("Given a binary, print the paths of dynamic library dependencies.");
    println!(
        "Given a script, print the paths of the dynamic library dependencies of the interpreter."
    );
    println!("Given -3, exit with success if the given binary is 32-bit.");
    println!("Given -6, exit with success if the given binary is 64-bit.");
    println!("Given -d, exit with success if the given binary is dynamically linked.");
    println!("Given -s, exit with success if the given binary is statically linked.");
    println!("Given -b, print the GNU build id of the file.");
    println!("Given -l, list the addresses and names of all symbols in the file.");
    println!("Given -n and a pattern, list the addresses and names of all symbols in the file whose name match the given pattern.");
    println!("Given -r and an address, resolve the name of the symbol at the given address.");
    println!("Given -t, print the type of the file.");
    println!("Given -v, print the version of the file.");
}

==> syd-3.22.0.tar.gz: syd-v3.22.0/src/syd-elf.rs <==
//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-elf.rs: syd's secure ldd(1) alternative
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    io::{Read, Seek, SeekFrom},
    path::{Path, PathBuf},
    process::ExitCode,
};

use anyhow::{bail, Context, Result};
use getargs::{Opt, Options};
use goblin::elf::{
    header::{ET_CORE, ET_DYN, ET_EXEC, ET_NONE, ET_REL},
    Elf,
};
use nix::unistd::Pid;
use regex::Regex;

fn main() -> Result<ExitCode> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = std::env::args().skip(1).collect::<Vec<_>>();

    // If no arguments are provided, display help
    if args.is_empty() {
        help();
        return Ok(ExitCode::SUCCESS);
    }

    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut root = None;
    let mut addr = None;
    let mut name = None;
    let mut list = false;
    let mut etyp = false;
    let mut ever = false;
    let mut build_id = false;
    let mut is_32bit = false;
    let mut is_64bit = false;
    let mut is_dynamic = false;
    let mut is_static = false;

    // SAFETY: We panic on parse errors.
    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('3') => {
                is_32bit = true;
            }
            Opt::Short('6') => {
                is_64bit = true;
            }
            Opt::Short('d') => {
                is_dynamic = true;
            }
            Opt::Short('s') => {
                is_static = true;
            }
            Opt::Short('b') => {
                build_id = true;
            }
            Opt::Short('l') => {
                list = true;
            }
            Opt::Short('t') => {
                etyp = true;
            }
            Opt::Short('v') => {
                ever = true;
            }
            Opt::Short('r') => {
                addr = match opts.value() {
                    Ok(value) => {
                        match u64::from_str_radix(
                            value.to_ascii_lowercase().trim_start_matches("0x"),
                            16,
                        ) {
                            Ok(value) => Some(value),
                            Err(_) => bail!("Invalid address specified for -r!"),
                        }
                    }
                    Err(_) => bail!("-r requires an argument!"),
                }
            }
            Opt::Short('n') => {
                name = match opts.value() {
                    Ok(value) => match Regex::new(value) {
                        Ok(value) => Some(value),
                        Err(_) => bail!("Invalid regular expression specified for -n!"),
                    },
                    Err(_) => bail!("-n requires an argument!"),
                }
            }
            Opt::Short('R') => {
                root = match opts.value() {
                    Ok(value) => Some(PathBuf::from(value)),
                    Err(_) => bail!("-R requires an argument!"),
                }
            }
            _ => bail!("Unknown option: {opt:?}!"),
        }
    }

    let flags = [
        build_id,
        is_32bit,
        is_64bit,
        is_dynamic,
        is_static,
        list,
        addr.is_some(),
        name.is_some(),
        etyp,
        ever,
    ];
    let parse = match flags.iter().filter(|&&flag| flag).count() {
        0 => false,
        1 => true,
        _ => bail!("At most one of -3, -6, -b, -d, -l, -n, -r, -s, t, and -v must be given!"),
    };

    let path = match opts.positionals().next() {
        None => bail!("Expected an ELF file as the first argument!"),
        Some(value) => Path::new(value)
            .canonicalize()
            .map_err(anyhow::Error::msg)
            .context("Failed to canonicalize argument!")?,
    };

    let pid = Pid::this();
    let (mut exe, size) = syd::fs::safe_open_file(&path)
        .map_err(anyhow::Error::msg)
        .context("Failed to open file for reading!")?;
    let prog = syd::fs::binfmt_script(&mut exe, pid, true)
        .map_err(anyhow::Error::msg)
        .context("Failed to determine interpreter")?;

    if parse {
        let bytes = if let Some(prog) = prog {
            if etyp {
                println!("script");
                return Ok(ExitCode::SUCCESS);
            }
            drop(exe);
            std::fs::read(prog)
                .map_err(anyhow::Error::msg)
                .context("Failed to read ELF file")?
        } else {
            exe.seek(SeekFrom::Start(0))
                .map_err(anyhow::Error::msg)
                .context("Failed to seek ELF file")?;

            let mut bytes = Vec::new();
            exe.read_to_end(&mut bytes)
                .map_err(anyhow::Error::msg)
                .context("Failed to read ELF file")?;
            drop(exe);
            bytes
        };
        let elf = Elf::parse(&bytes)
            .map_err(anyhow::Error::msg)
            .context("Failed to parse ELF file")?;

        if is_32bit {
            return Ok(if !elf.is_64 {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if is_64bit {
            return Ok(if elf.is_64 {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if is_dynamic {
            return Ok(if elf.dynamic.is_some() {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if is_static {
            return Ok(if elf.dynamic.is_none() {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if build_id {
            // Use `iter_note_headers` to iterate over PT_NOTE program headers.
            if let Some(notes) = elf.iter_note_headers(&bytes) {
                for note in notes.flatten() {
                    if note.name == "GNU" && note.n_type == 3 {
                        println!("{}", const_hex::encode(note.desc));
                        return Ok(ExitCode::SUCCESS);
                    }
                }
            }
            return Ok(ExitCode::FAILURE);
        } else if list {
            for sym in &elf.syms {
                if let Some(name) = elf.strtab.get_at(sym.st_name) {
                    println!("{:#x}:{name}", sym.st_value);
                }
            }
            return Ok(ExitCode::SUCCESS);
        } else if etyp {
            let name = match elf.header.e_type {
                ET_NONE => "none".to_string(),
                ET_REL => "rel".to_string(),
                ET_EXEC => "exe".to_string(),
                ET_DYN => "obj".to_string(),
                ET_CORE => "core".to_string(),
                _ => format!("{:#x}", elf.header.e_type),
            };
            println!("{name}");
            return Ok(ExitCode::SUCCESS);
        } else if ever {
            println!("{}", elf.header.e_version);
            return Ok(ExitCode::SUCCESS);
        } else if let Some(addr) = addr {
            let name = elf
                .syms
                .iter()
                .find(|sym| sym.st_value == addr)
                .and_then(|sym| elf.strtab.get_at(sym.st_name));
            return Ok(if let Some(name) = name {
                println!("{addr:#x}:{name}");
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else if let Some(regex) = name {
            let mut is_match = false;
            for sym in &elf.syms {
                if let Some(name) = elf.strtab.get_at(sym.st_name) {
                    if regex.is_match(name) {
                        is_match = true;
                        println!("{:#x}:{name}", sym.st_value);
                    }
                }
            }
            return Ok(if is_match {
                ExitCode::SUCCESS
            } else {
                ExitCode::FAILURE
            });
        } else {
            unreachable!();
        }
    }

    let deps = if let Some(ref p) = prog {
        drop(exe); // no longer necessary.

        let mut deps = vec![p.to_path_buf(), path];
        let (mut exe, size) = syd::fs::safe_open_file(p)
            .map_err(anyhow::Error::msg)
            .context("Failed to open interpreter")?;
        syd::fs::ldd(
            &mut exe, size, &mut deps, root, pid, false, false, false, true,
        )
        .map_err(anyhow::Error::msg)
        .context("Failed to analyze interpreter")?;
        drop(exe);

        deps
    } else {
        exe.seek(SeekFrom::Start(0))
            .map_err(anyhow::Error::msg)
            .context("Failed to seek ELF file")?;

        let mut deps = vec![path];
        syd::fs::ldd(
            &mut exe, size, &mut deps, root, pid, false, false, false, true,
        )
        .map_err(anyhow::Error::msg)
        .context("Failed to analyze ELF file")?;
        drop(exe);

        deps
    };

    for path in deps {
        println!("{}", path.display());
    }

    Ok(ExitCode::SUCCESS)
}

fn help() {
    println!("Usage: syd-elf [-36bdhlrstv] [-R root] binary|script");
    println!("Given a binary, print the paths of dynamic library dependencies.");
    println!(
        "Given a script, print the paths of the dynamic library dependencies of the interpreter."
    );
    println!("Given -3, exit with success if the given binary is 32-bit.");
    println!("Given -6, exit with success if the given binary is 64-bit.");
    println!("Given -d, exit with success if the given binary is dynamically linked.");
    println!("Given -s, exit with success if the given binary is statically linked.");
    println!("Given -b, print the GNU build id of the file.");
    println!("Given -l, list the addresses and names of all symbols in the file.");
    println!("Given -n and a pattern, list the addresses and names of all symbols in the file whose name match the given pattern.");
    println!("Given -r and an address, resolve the name of the symbol at the given address.");
    println!("Given -t, print the type of the file.");
    println!("Given -v, print the version of the file.");
}
